module TransfoChains where

import PrettyPrint
import PosSyntax
import AbstractIO
import Maybe
import TypedIds
import UniqueNames hiding (srcLoc)
import PNT
import TiPNT
import Data.List
import RefacUtils hiding (getParams) 
import PFE0 (findFile, epput)
import MUtils (( # ))
import qualified RefacLocUtils
import System
import IO
import qualified RefacNewDef
import RefacGenDef
import RefacMoveDef
import RefacMvDefBtwMod
import qualified RefacFunDef -- for subFunctionDef : Fold Definition
import RefacRenaming
import RefacDupDef
import RefacRmDef
import qualified RefacCleanImps
import qualified RefacGenFold
import TransfoUtils -- Part of the transformer layer
import qualified HsLexerPos -- for duplicateIntoComment

import qualified TransfoOperations
import qualified RefacSimplifyCase


new_def_branches funname prefix modname =
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case find_toplevel_definition_in_module funname mod of
        Just (Dec (HsFunBind _ matchlist)) -> 
            let poslist = map (\(HsMatch _ _ _ rhs _) -> RefacLocUtils.startEndLoc tokList rhs) matchlist
                poslist' = put_numbers 0 poslist
                actionstodo = map (\(((l,c),(l',c')),n) ->  RefacNewDef.introNewDef [fic,prefix ++ show n, show l, show c, show l', show c']) poslist'
            in sequence_ (reverse actionstodo) -- sequence_ executes the monads in its parameter /or transforms a list of monad into the monad for their sequence
             {- Important : there is a trick here. The list has to be reversed. If not, once the first case of the function definition is refactored, all the coordinates of the following cases become invalid. Refactoring from end of the definition to the top allows to avoid to parse the file several times to grep to correct coordiates. -}
        Nothing -> error ("No top-level declaration of " ++ funname ++ " found in " ++ modname ++ ".")
        Just _ -> error ("The definition of " ++ funname ++ " in " ++ modname ++ " must be a function (funbind).")


{- here, unlike in new_def_branches, the new prefix is the
   name of the function ++ "_" and the new suffix is the
   name of the first constructor in the patterns of the
   declaration (if any) -} 
new_def_branches2 funname modname =
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case find_toplevel_definition_in_module funname mod of
        Just (Dec (HsFunBind _ matchlist)) -> 
            let pos_constr_list = map (\(HsMatch _ _ pats rhs _) -> (   constructors_in_pats pats  ,RefacLocUtils.startEndLoc tokList rhs)) matchlist
                poslist' = put_numbers 0 pos_constr_list
                actionstodo = map (\((constr_list,((l,c),(l',c'))),n) -> 
                                       let name = make_name funname constr_list n in
                                       RefacNewDef.introNewDef [fic, name, show l, show c, show l', show c']) poslist'
            in sequence_ (reverse actionstodo) -- sequence_ executes the monads in its parameter /or transforms a list of monad into the monad for their sequence
             {- Important : there is a trick here. The list has to be reversed. If not, once the first case of the function definition is refactored, all the coordinates of the following cases become invalid. Refactoring from end of the definition to the top allows to avoid to parse the file several times to grep to correct coordiates. -}
        Nothing -> error ("No top-level declaration of " ++ funname ++ " found in " ++ modname ++ ".")
        Just _ -> error ("The definition of " ++ funname ++ " in " ++ modname ++ " must be a function (funbind).")

    where constructors_in_pats pats = mapMaybe get_constructor_of_pattern pats

          make_name funname [] n = funname ++ "_" ++ show n
          make_name funname ((c,_):_) _ = funname ++ "_" ++ c

{-
--FIXME : consider replaceing "Just 1" below by "Nothing"
generaliseAllRecAppToPatternVariablesInAllLocalDefsIn defname local_funs_prefix modname = 
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      let target_decls = find_target_decls_in_funbind_in_module defname local_funs_prefix mod
          l  = map (\d -> ((get_ident_PNT (extract_ident_of_fun_decl d)), (find_app_of defname (Just 1) Nothing d))) (target_decls::[HsDeclP]) -- FIXME : 100 -- contains pairs with a name of  a local decl and a list of applications of interest
          l' = concat (map (\(name,s) -> map (\ _ -> name) s) l) -- l' contains a number of occurences of name which correspond to the number of applications of defname to generalise in teh local definition of name
          l'' = put_numbers 0 l'
          actions = map (\(name,number) -> TransfoOperations.generalise_fun_app_in_in defname name defname ("x" ++ show number) modname) l''
          in if l == [] then error "generaliseAllRecAppToPatternVariablesInAllLocalDefsIn : no convenient application found."
             else sequence_ actions

-- attention, cette façon de faire avec un liste fixée avant de commencer les transformations peut poser problème lorsqu'un generalise va faire disparaître plusieurs instances identiques.
-}

--FIXME : consider replaceing "Just 1" below by "Nothing"
generaliseAllRecAppToPatternVariablesInAllLocalDefsIn defname local_funs_prefix modname = 
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      let target_decls = find_target_decls_in_funbind_in_module defname local_funs_prefix mod
          l            = map (find_app_of defname (Just 1) Nothing) target_decls
          l'           = map (put_numbers 0) l
          l''          = concat l'
      case l'' of 
        [] -> return ()
        _ -> let (target_app,num) = last l'' -- we take the last one in order to make the generalisations in a reverse order so that the arguments are introduced in a convenient order (FIXME: this is not always the best order)
                 ((l,c),(l',c')) =RefacLocUtils.startEndLoc tokList target_app
             in
               do 
                 RefacGenDef.generaliseDef [fic, "x" ++ show num, show l, show c, show l', show c'] -- FIXME: the new name is not very good
                 generaliseAllRecAppToPatternVariablesInAllLocalDefsIn defname local_funs_prefix modname


put_numbers _ [] = []
put_numbers n (e:r) = (e,n) : (put_numbers (n+1) r)



lift_branch_definitions_to_toplevel funname prefix modname =
    do filename<-PFE0.findFile (strToModName modname)
       (_, _, (mod::HsModuleP), tokList)<-parseSourceFile filename
       let target_decls = find_target_decls_in_funbind_in_module funname prefix mod
           target_positioned_names = map extract_ident_of_fun_decl target_decls 
           coords = map (RefacLocUtils.startEndLoc tokList) target_positioned_names 
           actionstodo = map (\((l,c),(_,_))-> liftToTopLevel [filename, show l, show c]) coords
           in sequence_ (reverse actionstodo)

-- generalise a given ident in all local declarations (with a given prefix) of a given toplevel declaration
generaliseIdentInAllLocalDefsIn defname local_funs_prefix targetname newname modname = 
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      let target_decls = find_target_decls_in_funbind_in_module defname local_funs_prefix mod
          -- In each declaration, the generalisation has to be done at most once.
          target_instances = mapMaybe (find_instance_of_ident_in_decl targetname) target_decls
          target_positions = map (RefacLocUtils.startEndLoc tokList) target_instances
          actions = reverse (map (\((l,c),(l',c')) -> RefacGenDef.generaliseDef [fic, newname, show l, show c, show l', show c']) target_positions)
          in sequence_ actions

-- generalise all idents (with a given property) in all toplevel declaration (with a given prorperty on the name) 
generaliseAllIdents (decl_name_prop::String->Bool) (target_name_prop::String->Bool) (newname_fun::(String,Int)->String) (modname::String) = 
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      let toplevel_decls = get_decl_list mod
          (target_decls::[HsDeclP]) = filter (decl_name_prop.get_ident_PNT.extract_ident_of_fun_pat_decl) toplevel_decls
          ident_decls = map (\d -> (d, nub (filter target_name_prop (map get_ident_PNT (extract_identifiers_decl d))))) target_decls
          (ident_decls_numbered::[ (HsDeclP, [(String,Int)])]) = map (\ (d,l) -> (d,put_numbers 0 l)) ident_decls
          ident_decl_pair = concat (map (\ (d,l) -> let n = get_ident_PNT (extract_ident_of_fun_pat_decl d) in map (\i -> (n,i)) l) ident_decls_numbered)
--          name_name_pair = nub (map (\(d,i) -> ( , get_ident_PNT i)) ident_decl_pair)
          actions = map ( \ (decl_name, (ident_name,i)) -> TransfoOperations.generaliseident (decl_name, modname) ident_name (newname_fun (ident_name,i))) ident_decl_pair

          in if ident_decl_pair == [] then error ("Nothing to generalise" ++ if target_decls == [] then " (no delaration selected)." else ".") else sequence_ (reverse actions)




-- Given a function funname, the parameters (in fact identifiers) of funcame are generalised in the local functions of funname which name begins by prefix.
generaliseAllPatternVariablesInAllLocalDefsIn funname prefix modname =
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case find_toplevel_definition_in_module funname mod of
        Just (Dec (HsFunBind _ matchlist)) -> 
            let l = map (\(HsMatch _ _ pats _ ds) -> (bound_variables_in_patterns pats,
                                                      filter (is_fun_pat_bind_with_prefix prefix) ds)) 
                        (matchlist::[HsMatchP])

                l' = map (\(pnts,ds) -> (reverse (map (get_ident_PNT) pnts), -- generalise put the new argument at the first position, so we have to reverse here to keep the same order as in the original patterns
                                         map (get_ident_PNT.extract_ident_of_fun_decl) ds)
                         ) l --we keep only names as coordinates will change after each refactoring
                l'' = flat l' -- pairs (list of names, name of a definition)
                in apply_all l''

    where flat [] = []
          flat (([],_):r)      = flat r
          flat ((names,[]):r)  = flat r
          flat ((names, (d:ds)):r) = (names,d) : (flat ( (names,ds) : r))
                  
          apply_all [] = return ()
          apply_all (([],_):r) = apply_all r
          apply_all ( ((name:names),d):r) = do TransfoOperations.generalise_ident_in_in funname d name name  modname ;
                                               apply_all ( (names,d):r)




moveAllDefsToConstructorModulesFollowingPatternMatching motherFunction businessFunctionPrefix sourceModule =
    do
      fic <- PFE0.findFile (strToModName sourceModule)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case find_business_function_at_toplevel mod businessFunctionPrefix tokList of
        Nothing -> return ()
        Just (name,((l,col),(_,_))) -> 
            case find_toplevel_definition_in_module motherFunction mod of
              Nothing -> error ("moveAllDefsToConstructorModulesFollowingPatternMatching : mother function not found.")
              Just d -> 
                  case find_pattern_constructor_for_case_using_ident d name of
                    Nothing -> error ("moveAllDefsToConstructorModulesFollowingPatternMatching : could not deduce the destination module from the pattern.")
                    Just con -> do 
                      RefacMvDefBtwMod.moveDefBtwMod [fic,(con++"Mod"),show l,show col] -- (con ++ "Mod") is very ad-hoc
                      moveAllDefsToConstructorModulesFollowingPatternMatching motherFunction businessFunctionPrefix sourceModule

    where find_business_function_at_toplevel mod prefix toklist = 
              let decls = get_decl_list mod in     
              case find (is_fun_pat_bind_with_prefix prefix) decls of
                Nothing -> Nothing
                Just d -> let i = extract_ident_of_fun_pat_decl d in  Just  (get_ident_PNT i, RefacLocUtils.startEndLoc toklist i)


          find_pattern_constructor_for_case_using_ident ::  HsDeclP -> String -> Maybe String
          find_pattern_constructor_for_case_using_ident (Dec (HsFunBind _ matches)) i = 
              case find_in_branch matches of 
                Nothing -> Nothing
                Just (HsMatch _ _ ps _ _) -> case constructors_in_patterns ps of
                                               [] -> Nothing
                                               pnt : _ -> Just (get_constructor_ident_PNT pnt)


              where
                find_in_branch [] = Nothing
                find_in_branch (m:ms) = case find_instance_of_ident_in_match i m of
                                          Nothing -> find_in_branch ms
                                          Just _ -> Just m


moveAllDefsOnSuffix funprefix modname =
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case find_toplevel_ident_prefixed_in_module funprefix mod of
        Nothing -> return ()
        Just pnt -> let suf = extract_suffix pnt
                        src = RefacUtils.useLoc pnt
                        destmodname = suf ++ "Mod"
                    in do RefacMvDefBtwMod.moveDefBtwMod [fic,destmodname ,show (srcLine src),show (srcColumn src)] -- (con ++ "Mod") is very ad-hoc
                          moveAllDefsOnSuffix funprefix modname

    where extract_suffix pnt  = 
              let name = get_ident_PNT pnt 
              in case Data.List.stripPrefix funprefix name of
                   Nothing -> error ("Cannot guess where to move " ++ name ++".")
                   Just s -> s

-- Find in "funname" all the identifiers starting with the given prefix 
-- and rename them into "newname". Qualifiers will be inserted if necessary (by HaRe).
-- all the new names are the same (they are defined in separate modules)            
collect_business_functions_and_rename funname businessFunctionPrefix newname modname =
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      let
          decls         = get_decl_list mod   
          (idents::[PNT])        = extract_identifiers_decls decls 
          target_idents = filter (has_prefix_PNT businessFunctionPrefix) idents 
          (infos::[(String,String)])         = remove_Nothings (nub (map (\ pnt -> (get_ident_PNT pnt, get_definition_module_PNT pnt)) target_idents))
 
      sequence_ (map (\ (i,m) -> TransfoOperations.rename_toplevel (i,m) newname) infos)
                 

    where remove_Nothings [] = []
          remove_Nothings ((i,Nothing):r) = remove_Nothings r
          remove_Nothings ((i,Just m):r)  = (i,m) : (remove_Nothings r)
             
-- Find in "funname" all the identifiers named "targetname" (possibly with different qualifiers)
-- and rename them with different names starting with the given prefix "newprefix".
-- All the new names are different (with a given prefix).
-- If "funname" and "targetname" are the same, the occurences refering to the toplevel function being handled will not be renamed.
renameAllIdentsIn2 funname targetname newprefix funmod c =
    do
      fic <- PFE0.findFile (strToModName funmod)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case find_toplevel_definition_in_module funname mod of
        Nothing -> error "function not found"
        Just d -> case find_instance_of_ident_in_decl_special targetname d of
                    Nothing -> return ()
                    Just i -> let m = case get_definition_module_PNT i of
                                        Nothing -> error "renameAllIdentIn: local definition : not implemented yet"
                                        Just m' -> m'
                              in
                                do 
                                  TransfoOperations.rename_toplevel (targetname, m) (newprefix ++ (show c));
                                  renameAllIdentsIn2 funname targetname newprefix funmod (c+1)
                                    
    where find_instance_of_ident_in_decl_special targetname d =
              let defined_pnt = extract_ident_of_fun_pat_decl d
              in  find (\ pnt -> (not (instance_of_PNT pnt defined_pnt)) && 
                                   is_ident_PNT targetname pnt) (extract_identifiers_decl d)
   
-- in the definition of modname.funname, unfold all identifiers with property p.    
allUnfoldIn p funname modname =
        do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case find_toplevel_definition_in_module funname mod of
        Nothing -> error "function not found"
        Just d -> let ids = extract_identifiers_decl d
                  in case find p ids of
                       Nothing -> return ()
                       Just pnt -> 
                           do
                             unfold_pnt pnt fic;
                             allUnfoldIn p funname modname

--FIXME : DO NOT call findFile at each recursive call

-- in modname, unfold all identifiers with property p.    
unfoldAll (p::PNT->Bool) modname = -- warning, the property must be stable with respect to changes in the given module
     do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      let ds = get_decl_list mod 
      case find p (extract_identifiers_decls ds) of
                       Nothing -> return ()
                       Just pnt -> 
                                    do
                                      unfold_pnt pnt fic;
                                      unfoldAll p modname

unfold_pnt pnt fic = 
    let src = RefacUtils.useLoc pnt in
    RefacNewDef.unfoldDef [fic, show (srcLine src), show (srcColumn src)]

-- no need to be in the interface because we want to unfold all instance in order to delete a definition after.
-- unfold all instances of modname.name (in all client modules also)
{-unfoldAllInstancesTopLevel name modname=     
 do
   fic <- PFE0.findFile (strToModName modname)
   (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
   case find_toplevel_definition_in_module name mod of
     Nothing -> error ("Identifier " ++ name ++ " not found at top level in this module (" ++ modname ++ ").")
     Just d -> let (ref_pnt::PNT) = extract_ident_of_fun_pat_decl d 
               in unfoldAllInstances_PNT ref_pnt modname
-}

{-unfoldAllInstancesAllIdentsTopLevel_prop (p::PNT->Bool) modname = 
 do
   fic <- PFE0.findFile (strToModName modname)
   (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
   case find p (map extract_ident_of_fun_pat_decl (get_decl_list mod)) of
     Nothing -> error ("No identifier with the given property found at top level in this module (" ++ modname ++ ").")
     Just i -> unfoldAllInstances_PNT i modname
-}
unfoldAllInstances_PNT_in_client_mods (ref_pnt::PNT) modname =
    let is_occurence p = TransfoUtils.instance_of_PNT p ref_pnt
    in
      do
        clients {- ::[(HsName.ModuleName, String)] -} <- clientModsAndFiles (strToModName modname)
        sequence_ (map 
                     (\ (modname,_) -> unfoldAll is_occurence (modNameToStr modname) )
                     clients
                  )

-- Unfold all the occurence of toplevel declaration with a given property in a given module.
-- The unfolded instances and the declarations are in a single module.
unfoldAllInstancesTopLevel_in_defining_mod (name_prop::String->Bool) modname =
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      let decls = get_decl_list mod
      case find (name_prop.get_ident_PNT) (defined_var_pnts decls) of
        Nothing -> error "No declaration with the given property."
        Just ref_pnt -> let is_occurence p = TransfoUtils.instance_of_PNT p ref_pnt
                        in case find is_occurence (extract_identifiers_decls decls) 
                           of Nothing -> return () --no instance left in this module
                              Just pnt -> do
                                unfold_pnt pnt fic ;
                                unfoldAllInstancesTopLevel_in_defining_mod name_prop modname

--move this decl (demote in forgetTopLevel?)
defined_var_pnts::[HsDeclP]->[PNT]
defined_var_pnts decls = map extract_ident_of_fun_pat_decl (filter is_fun_pat_bind decls)

forgetTopLevel name_prop modname = 
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      let decls = get_decl_list mod
          pnts = filter (name_prop.get_ident_PNT) (defined_var_pnts decls)
      if pnts == [] 
         then error "No declaration found"
         else let actions_for_client_mods = map (\pnt -> unfoldAllInstances_PNT_in_client_mods pnt modname) pnts
                  actions_for_removing_exports = map (\pnt -> TransfoOperations.rmFromExports (get_ident_PNT pnt) modname) pnts
                  actions_for_removing    = map (\pnt -> TransfoOperations.removedef (get_ident_PNT pnt) modname) pnts
              in do
                 sequence_ actions_for_client_mods ;
                 unfoldAllInstancesTopLevel_in_defining_mod name_prop modname ;
                 sequence_ actions_for_removing_exports
                 sequence_ actions_for_removing 

forgetIn funname localname modname = 
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case find_toplevel_definition_in_module funname mod of
        Nothing -> error "function not found"
        Just d -> case find_definition localname [d] of
                        Just d' ->
                            let 
                                defpnt =  extract_ident_of_fun_pat_decl d'
                                is_occurence p = TransfoUtils.instance_of_PNT p defpnt  
                            in case find is_occurence  (extract_identifiers_decl d) of
                                 Just instance_pnt -> do 
                                   unfold_pnt instance_pnt fic ;
                                   forgetIn funname localname modname
                                 Nothing -> TransfoOperations.removelocaldef localname funname modname
                        Nothing -> error "local def not found"


forgetTopLevelIdentWithOccurenceIn funname name_prop modname = 
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case find_toplevel_definition_in_module funname mod of
        Nothing -> error "function not found"
        Just d -> case find (name_prop.get_ident_PNT) (extract_identifiers_decl d) of
                    Nothing -> error "No declaration with the given property."
                    Just pnt -> forget_toplevel_PNT pnt

forgetAllTopLevelIdentWithOccurenceIn funname name_prop modname = 
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case find_toplevel_definition_in_module funname mod of
        Nothing -> error "function not found"
        Just d -> case find (name_prop.get_ident_PNT) (extract_identifiers_decl d) of
                    Nothing ->  return ()
                    Just pnt -> do 
                                 forget_toplevel_PNT pnt;
                                 forgetAllTopLevelIdentWithOccurenceIn funname name_prop modname

-- forget a PNT pointing to an occurence of a global declaration (unfold all the instances of the same decl and delete this decl)
forget_toplevel_PNT pnt = 
    case get_definition_module_PNT pnt of
      Nothing -> error "Unable to locate the toplevel def fot the target ident."
      Just m -> let name = get_ident_PNT pnt in 
                forgetTopLevel ((==) name) m


                             
generative_fold_all_ident_app_in funname identapplied modname = 
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case find_toplevel_definition_in_module funname mod of
        Nothing -> error "function not found"
        Just d -> case find_app_of identapplied Nothing Nothing d of
                    [] -> return ()
                    e:_ -> do TransfoOperations.generativeFold e modname ;
                              generative_fold_all_ident_app_in funname identapplied modname



simplifyCasePatternInRepeat funname modname =
     do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, mod, tokList)<-parseSourceFile fic
      case find_toplevel_definition_in_module funname mod of
        Nothing -> error "function not found"
        Just d -> case TransfoUtils.select_case d of
                     [] -> error ("no case found in  " ++ funname ++".")
                     case_expr : _ ->
                         case RefacSimplifyCase.simplify  tokList case_expr of
                           Nothing -> return ()
                           Just (id,tokList') -> 
                               do 
                                 writeRefactoredFiles False [((fic,RefacLocUtils.modified),(tokList',mod))] ; -- mod is unused in that function
                                 forgetIn funname id modname ;
                                 simplifyCasePatternInRepeat funname modname