{- Julien Cohen 2010-2011 -}
module TransfoOperations where

import PrettyPrint
import PosSyntax
import AbstractIO
import Maybe
import TypedIds
import UniqueNames hiding (srcLoc)
import SrcLoc1 (loc0)
import PNT
import TiPNT
import List
import RefacUtils hiding (getParams) 
import PFE0 (findFile, epput)
import MUtils (( # ))
import qualified RefacLocUtils
import System
import IO
import qualified RefacNewDef
import RefacGenDef
import RefacMoveDef
import RefacMvDefBtwMod
import qualified RefacFunDef -- for subFunctionDef : Fold Definition
import RefacRenaming
-- import qualified RefacUnsafeRenaming
import RefacDupDef
import RefacRmDef
import qualified RefacCleanImps
import qualified RefacGenFold
import TransfoUtils -- Part of the transformer layer
import qualified HsLexerPos -- for duplicateInotComment
import qualified RefacAlphaEq


-- Transform an argument of fold into a local binding

local_definition_of_fun_arg def_name arg_pos new_name modname = 
    do 
      filename<-PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile filename 
      let ((a,b),(c,d)) = coordinates_of_app mod tokList def_name arg_pos
      RefacNewDef.introNewDef [filename, new_name, show a , show b, show c, show d] 
 
     
-- transform a local declaration into a global one

make_global_of_local local_name modname = -- warning : only a the level directly under the top-level -- this should have been fixed
    do filename<-PFE0.findFile (strToModName modname)
       (_, _, (mod::HsModuleP), tokList)<-parseSourceFile filename
       let ((l,c),(_,_)) = coordinates_of_local_definition_identifier mod tokList local_name 
       liftToTopLevel [filename, show l, show c]


make_global_of_local_in fun_name local_name modname =
    do filename<-PFE0.findFile (strToModName modname)
       (_, _, (mod::HsModuleP), tokList)<-parseSourceFile filename
       let ((l,c),(_,_)) = coordinates_of_local_definition_identifier_in mod tokList fun_name local_name 
       liftToTopLevel [filename, show l, show c]

lift_one_level_in fun_name local_name modname =
    do filename<-PFE0.findFile (strToModName modname)
       (_, _, (mod::HsModuleP), tokList)<-parseSourceFile filename
       let ((l,c),(_,_)) = coordinates_of_local_definition_identifier_in mod tokList fun_name local_name 
       liftOneLevel [filename, show l, show c]

demote_toplevel (name::String) srcmodname  =
    do 
      fic <- PFE0.findFile (strToModName srcmodname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case coordinates_of_global_definition_identifier mod tokList name of 
        Just ((l,c),(_,_)) -> RefacMoveDef.demote[fic,show l,show c]
        Nothing -> error ("No top-level declaration of " ++ name ++ " found in " ++ srcmodname ++ ".")

demote_in (name::String) funname srcmodname  =
    do 
      fic <- PFE0.findFile (strToModName srcmodname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case coordinates_of_local_definition_identifier_in mod tokList funname name of 
        {-Just-} ((l,c),(_,_)) -> RefacMoveDef.demote[fic,show l,show c]
        {-Nothing -> error ("No local declaration of " ++ name ++ " found in " ++ funname ++ ".")-}

-- move a definition between modules

move_def_between_modules (name::String) srcmodname destmodname =
    do 
      fic <- PFE0.findFile (strToModName srcmodname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case coordinates_of_global_definition_identifier mod tokList name of 
        Just ((l,c),(_,_)) -> moveDefBtwMod[fic,destmodname,show l,show c]
        Nothing -> error ("No top-level declaration of " ++ name ++ " found in " ++ srcmodname ++ ".")
      

-- Renamings 

rename_toplevel (oldname,modname) newname  =
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case coordinates_of_global_definition_identifier mod tokList oldname of 
        Just ((l,c),(_,_)) -> RefacRenaming.rename [fic,newname, show l, show c]
        Nothing -> error ("No top-level declaration of " ++ oldname ++ " found in " ++ modname ++ ".")

{- rename_toplevel_unsafe (oldname,modname) newname  =
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case coordinates_of_global_definition_identifier mod tokList oldname of 
        Just ((l,c),(_,_)) -> RefacUnsafeRenaming.rename_unsafe [fic,newname, show l, show c]
        Nothing -> error ("No top-level declaration of " ++ oldname ++ " found in " ++ modname ++ ".")
-}

rename_local (oldname,modname) newname  = 
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      let ((l,c),(_,_)) = coordinates_of_local_definition_identifier mod tokList oldname
      RefacRenaming.rename [fic,newname, show l, show c]


-- exhibit the functions in a patterned definition


exhibit_function funname constrname newname modname =
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      let ((l,c),(l',c')) = coordinates_of_case_body mod tokList funname constrname
      RefacNewDef.introNewDef [fic,newname , show l, show c, show l', show c']


-- other_type means that the generalization is done on a value which type is not the type of interest
-- funname is the name of the function of interest
-- constrname is the name of the constructor which case has to be handled
-- modname is the module where the function is defined
-- pos si the position of the argument of the constructor to consider (0 for the first one)
-- newargname is the name of the new parameter after generalization
-- curryflag indicates if the arguments of contr are curried or not
generalisePatternVariableInLocalDefIn funname constrname localfunname modname pos (curryflag::CurryMode)  = 
    generalisePatternVariableInLocalDefIn2 funname constrname localfunname modname pos curryflag 0

generalisePatternVariableInLocalDefIn2 funname constrname localfunname modname pos (curryflag::CurryMode) patternposition= 
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      let (((l,c),(l',c')), varname) = coordinates_of_instance_of_pos_constr_arguments mod tokList funname constrname pos localfunname curryflag  patternposition
      RefacGenDef.generaliseDef [fic,varname, show l, show c, show l', show c']


generaliseRecAppToPatternVariableInLocalDefIn funname constrname localfunname modname pos newargname (curryflag::CurryMode)  = 
    generaliseRecAppToPatternVariableInLocalDefIn2 funname constrname localfunname modname pos newargname curryflag 0 

generaliseRecAppToPatternVariableInLocalDefIn2 funname constrname localfunname modname pos newargname (curryflag::CurryMode) patternposition= 
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      let ((l,c),(l',c')) = coordinates_of_rec_call_applied_to_instance_of_pos_constr_arguments mod tokList funname constrname pos localfunname curryflag  patternposition
      RefacGenDef.generaliseDef [fic,newargname, show l, show c, show l', show c']


-- it is nearly a copy-paste of new_def_fun_app_in
-- optFirstArgName is an optional string to specify the name of the first argument (a variable)
generalise_fun_app_in defname funname optnbargs optFirstArgName newname modname =
  do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case find_definition defname (get_decl_list mod) of
        Nothing -> error ("No declaration of " ++ defname ++ " found in module " ++ modname ++ ".")
        Just d ->
            case find_app_of funname optnbargs optFirstArgName d of
              [] -> let optmessage = case optFirstArgName of Nothing -> ""
                                                             Just s -> " and to " ++ s 
                    in error ("No application of " ++ funname ++ " to " ++ show optnbargs ++ " arguments " ++ optmessage ++ " found in module " ++ modname ++ " in declaration of " ++ defname ++".")
              (a:_) -> let ((l,c),(l',c')) =RefacLocUtils.startEndLoc tokList a
                       in RefacGenDef.generaliseDef [fic,newname, show l, show c, show l', show c']

generalise_fun_arg_in defname funname argName newname modname =
  do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case find_definition defname (get_decl_list mod) of
        Nothing -> error ("No declaration of " ++ defname ++ " found in module " ++ modname ++ ".")
        Just d -> case find_app_of funname Nothing Nothing d of
                    [] -> error ("generalise_fun_arg_in: No application of " ++ funname ++ " in the declaration of " ++ defname ++ " found in module " ++ modname ++ ".")
                    apps -> case find_argument_in_app apps
                            of [] -> error ("generalise_fun_arg_in: No instance of " ++ argName ++" found in an application of " ++ funname ++ " in the declaration of " ++ defname ++ " in module " ++ modname ++ ".")
                               pnt:_ -> 
                                   let ((l,c),(l',c')) =RefacLocUtils.startEndLoc tokList pnt
                                   in RefacGenDef.generaliseDef [fic,newname, show l, show c, show l', show c']

  where 

    find_argument_in_app exps = mapMaybe (find_instance_of_ident_in_expr argName) exps


-- not exported to emacs ,  used in generaliseRecCallsInBranches
generalise_fun_app_in_in defname localdefname funname {-nbargs optFirstArgName-} newname modname =
  do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case find_definition defname (get_decl_list mod) of
        Nothing -> error ("No declaration of " ++ defname ++ " found in module " ++ modname ++ ".")
        Just d -> let ds = get_all_decl_next_dec d 
                      target_ds = filter (TransfoUtils.defines localdefname) ds
                  in
                    case concat (map (find_app_of funname Nothing {- 1 -} Nothing) target_ds) of -- FIXME : 100
                      [] -> error ("No application of " ++ funname ++ " found in module " ++ modname ++ " in declaration of " ++ defname ++".")
                      (a:_) -> let ((l,c),(l',c')) =RefacLocUtils.startEndLoc tokList a
                               in   RefacGenDef.generaliseDef [fic,newname, show l, show c, show l', show c']


--Geneneralise target ident in the toplevel definition of modname.funname
generaliseident (funname,modname) targetident newname =
 do 
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic


      case coordinates_of_instance mod tokList (Just funname) targetident of
        Just ((l,c),(l',c')) -> RefacGenDef.generaliseDef [fic,newname, show l, show c, show l', show c'] 
        Nothing -> error ("no instance of " ++ targetident ++ " found in " ++ funname ++ ".")

-- not exported to Emacs
-- Used in TransfoChains.generaliseAllPatternVariablesInAllLocalDefsIn
generalise_ident_in_in defname localdefname ident newname modname =
  do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case find_definition defname (get_decl_list mod) of
        Nothing -> error ("No declaration of " ++ defname ++ " found in module " ++ modname ++ ".")
        Just d -> let ds = get_all_decl_next_dec d
                  in 
                      case find (TransfoUtils.defines localdefname) ds of
                        Nothing -> error ("No declaration of " ++ localdefname ++ " found in " ++ defname ++ ".") -- TODO : faire pareil dans generalise_fun_app_in_in
                        Just d ->
                            case  find_instance_of_ident_in_decl ident d of 
                              Nothing -> return ()
                              Just a -> 
                                  let ((l,c),(l',c')) = RefacLocUtils.startEndLoc tokList a
                                  in  RefacGenDef.generaliseDef [fic,newname, show l, show c, show l', show c']


doduplicate (funname,modname) newname =
    do 
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case coordinates_of_global_definition_identifier mod tokList funname of
         Just ((l,c),(_,_)) -> duplicateDef [fic,newname, show l, show c]
         Nothing -> error ("No top-level declaration of " ++ funname ++ " found in " ++ modname ++ ".")


unfoldinstance id modname =
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case coordinates_of_instance mod tokList Nothing id of
        Just ((l,c),(_,_)) -> RefacNewDef.unfoldDef [fic, show l, show c] 
        Nothing -> error ("no instance of " ++ id ++ " found in " ++ modname ++ ".")

unfoldinstance_in id_to_unfold id_decl_context modname =
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case coordinates_of_instance mod tokList (Just id_decl_context) id_to_unfold of
        Just ((l,c),(_,_)) -> RefacNewDef.unfoldDef [fic, show l, show c] 
        Nothing -> error ("no instance of " ++ id_to_unfold ++ " found in " ++ modname ++ ".")

fold_toplevel_definition id modname =
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case coordinates_of_toplevel_declaration mod tokList id of
        Just ((l,c),(l',c')) -> RefacFunDef.subFunctionDef [fic, show l, show c, show l', show c'] 
        Nothing -> error ("No toplevel declaration of " ++ id ++ " found in " ++ modname ++ ".")
 

-- optFirstArgName is an optional string to specify the name of the first argument (a variable)
new_def_fun_app funname optNbargs optFirstArg newname modname =
  do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      let bs =  extract_all_bodies_of_module mod
          apps = concat (map (get_apps funname optFirstArg) bs) 
      case apps of 
        [] -> error ("no application of " ++ funname ++ " found in " ++ modname ++ ".")
        a:r  ->  -- une seule fois devrait suffire, pas d'appel récursif, 
                 -- la nouvelle définition devrait se faire à plusieurs endroits automatiquement, si nécessaire en utilisant un folding
          let ((l,c),(l',c')) = 
                case optNbargs of
                  Just nbargs ->
                      case extract_app_to_n_args nbargs a of
                        Just a' ->  RefacLocUtils.startEndLoc tokList a'
                        Nothing -> error ("No application to " ++ show nbargs ++ " found.")
                  Nothing -> RefacLocUtils.startEndLoc tokList  a
            in RefacNewDef.introNewDef [fic,newname, show l, show c, show l', show c']



-- optFirstArgName is an optional string to specify the name of the first argument (a variable)
new_def_fun_app_in defname funname nbargs optFirstArgName newname modname =
  do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case find_definition defname (get_decl_list mod) of
        Nothing -> error ("No declaration of " ++ defname ++ " found in module " ++ modname ++ ".")
        Just d -> 
            case find_app_of funname nbargs optFirstArgName d of
              [] -> let optmessage = case optFirstArgName of Nothing -> ""
                                                             Just s -> " and to " ++ s 
                    in error ("No application of " ++ funname ++ " to " ++ show nbargs ++ " arguments " ++ optmessage ++ " found in module " ++ modname ++ " in declaration of " ++ defname ++".")
              (a:_) ->  let ((l,c),(l',c')) =RefacLocUtils.startEndLoc tokList a
                        in RefacNewDef.introNewDef [fic,newname, show l, show c, show l', show c']

new_def_infix_app_in defname funname newname modname =
  do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case find_definition defname (get_decl_list mod) of
        Nothing -> error ("no declaration of " ++ defname ++ " found in module " ++ modname ++ ".")
        Just d ->
            let bs =  extract_bodies_of_decl d
                apps = concat (map (get_infix_apps funname) bs) 
            in case apps of 
                         [] -> error ("no application of " ++ funname ++ "found in module " ++ modname ++ " in declaration of " ++ defname ++".")
                         a:r  ->  -- une seule fois devrait suffire, pas d'appel récursif, 
                                  -- la nouvelle définition devrait se faire à plusieurs endroits automatiquement, si nécessaire en utilisant un folding
                                   let ((l,c),(l',c')) =RefacLocUtils.startEndLoc tokList a
                                   in RefacNewDef.introNewDef [fic,newname, show l, show c, show l', show c']
      
new_def_ident_in defname ident newname modname =
  do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case find_definition defname (get_decl_list mod) of
        Nothing -> error ("no declaration of " ++ defname ++ " found in module " ++ modname ++ ".")
        Just d ->
            case find_instance_of_ident_in_decl ident d of
              Nothing -> error ("no instance of " ++ ident ++ " found in module " ++ modname ++ " in declaration of " ++ defname ++".")
              Just e ->
                    let ((l,c),(l',c')) = RefacLocUtils.startEndLoc tokList e
                    in RefacNewDef.introNewDef [fic,newname, show l, show c, show l', show c']
      

removedef name modname =   
  do
    fic <- PFE0.findFile (strToModName modname)
    (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
    case coordinates_of_global_definition_identifier mod tokList name of
      Just ((l,c),(_,_)) -> removeDef [fic, show l, show c]
      Nothing -> error ("No top-level declaration of " ++ name ++ " found in " ++ modname ++ ".")

removelocaldef localname funname modname =   
  do
    fic <- PFE0.findFile (strToModName modname)
    (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
    let ((l,c),(_,_)) = coordinates_of_local_definition_identifier_in mod tokList funname localname
    removeDef [fic, show l, show c]



duplicateIntoComment name modname = 
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case find (TransfoUtils.defines name) (get_decl_list mod) of
        Nothing -> error ("No declaration for " ++ show name ++ " found in " ++ show modname ++ ".")
        Just d -> do let (s,_)   = RefacLocUtils.startEndLoc  tokList d 
                         newtoks = RefacLocUtils.insertComments (s,s) tokList ({-RefacLocUtils. ? -} prettyprint d) 
                     writeRefactoredFiles False [((fic,RefacLocUtils.modified),(newtoks,mod))]


rmCommentBefore name modname = 
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case find (TransfoUtils.defines name) (get_decl_list mod) of
        Nothing -> error ("No declaration for " ++ show name ++ " found in " ++ show modname ++ ".")
        Just d ->    let (s,_)   = RefacLocUtils.startEndLoc  tokList d 
                         (a,b,c) = RefacLocUtils.splitToks (s,s) tokList 
                         a' = reverse a
                         a'' = reverse (removeComment a')
                     in writeRefactoredFiles False [((fic,RefacLocUtils.modified),(a''++b++c,mod))]

    where removeComment [] = []
          removeComment ( (RefacLocUtils.Whitespace, _) : r) = removeComment r
          removeComment ( (RefacLocUtils.Comment, _)    : r) = r
          removeComment ( (RefacLocUtils.NestedComment, _)    : r) = r
          removeComment t = error ("Don't know what to do with token " ++ show t ++ " while removing comments.")

--
select_app_of_fun_to_n_args fun_name nbargs modname =
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), _)<-parseSourceFile fic

      case TransfoUtils.select_apps_of_fun_to_n_args fun_name nbargs mod of
        [] -> error ("no application of " ++ fun_name ++ " to " ++ (show nbargs) ++" args found.")
        a:_ -> return a
       
select_case_in fun_name modname = 
     do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), _)<-parseSourceFile fic

      case TransfoUtils.find_toplevel_definition_in_module fun_name mod of
        Nothing -> error ("no declaration of " ++ fun_name ++"found.")
        Just d  -> case TransfoUtils.select_case d of
                     [] -> error ("no case found in  " ++ fun_name ++".")
                     a:_ -> return a



select_toplevel_decl fun_name modname =
     do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), _)<-parseSourceFile fic

      case TransfoUtils.find_toplevel_definition_in_module fun_name mod of
        Nothing -> error ("No declaration of  " ++ fun_name ++" found.")
        Just a -> return a

generativeFold exp modname = 
    do fic <- PFE0.findFile (strToModName modname)
       (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
       let ((l,c),(l',c')) =RefacLocUtils.startEndLoc tokList exp
       RefacGenFold.genFold [fic, show l, show c, show l', show c']


cleanImports modname =
    do fic <- PFE0.findFile (strToModName modname)
       RefacCleanImps.cleanImports [fic]

rmFromExports funname modname =
    do fic <- PFE0.findFile (strToModName modname)
       (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
       case coordinates_of_ident_in_exports mod tokList funname of
         Nothing -> return ()
         Just ((l,c),(_,_)) -> RefacCleanImps.rmFromExport [fic, show l, show c]


addImplicitImport_unsafe in_modname import_modname = 
    do in_fic     <- PFE0.findFile (strToModName in_modname)
       import_fic <- PFE0.findFile (strToModName import_modname)
       (_, _, (in_mod::HsModuleP), in_tokList)<-parseSourceFile in_fic
       (_, _, (import_mod::HsModuleP), import_tokList)<-parseSourceFile import_fic
       (in_free,in_declared) <- RefacUtils.hsFDNamesFromInside    in_mod -- it is not sufficient because some names could be visible from imports and exported.
       ( _, imp_exported)    <- RefacUtils.hsFreeAndDeclaredNames import_mod 
       case is_in_imports import_modname (hsModImports in_mod)  of
            Nothing -> let needs_qualified = clash in_free imp_exported || clash in_declared imp_exported
                       in add_import needs_qualified import_modname in_mod in_tokList in_fic
            Just Nothing -> return ()
            Just (Just _) -> error "Transform an explicit import into an implicit import : not implemented."
    where
      add_import needs_qualified import_modname (in_mod_ast::HsModuleP) in_toklist in_fic =
          let module_offset = RefacUtils.moduleOffset in_mod_ast in_toklist
              directive = HsImportDecl SrcLoc1.loc0 (SN (strToModName import_modname) SrcLoc1.loc0) needs_qualified Nothing {- no as -} Nothing {- no explicit imports -}
              directiveToks = RefacLocUtils.tokenise HsLexerPos.startPos module_offset False ("\n" ++ RefacLocUtils.prettyprint (directive::HsImportDeclP) ++"\n\n") 
          in let newtoks = case hsModDecls in_mod_ast of
                             [] -> in_toklist ++ directiveToks
                             d:_ -> 
                                 let (pos,_) = getStartEndLoc  in_toklist d
                                     (toks1, dtoks, toks2) = RefacLocUtils.splitToks (pos,pos) in_toklist
                                    in toks1 ++ directiveToks ++ dtoks ++ toks2
             in writeRefactoredFiles False [((in_fic,RefacLocUtils.modified),(newtoks,(in_mod_ast::HsModuleP)))] -- in_mod_ast should be unused


      clash l1 l2 = case intersect l1 l2 of
                      [] -> False
                      _:_ -> True -- I hope I don't need to compute the whole intersection if it is non empty because of lazyness.

      is_in_imports modname [] = Nothing
      is_in_imports modname ( (HsImportDecl _ (SN name _) qualified_flag _ explicit_imports) : r) =
          if (modNameToStr name) == modname
          then Just explicit_imports
          else is_in_imports modname r


-- Replace alpha-equivalent idents

replaceIdentIn ident decl_name modname target_ident =
     do fic <- PFE0.findFile (strToModName modname)
        (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
        let d = case find_toplevel_definition_in_module decl_name mod of
                  Just res -> res
                  Nothing -> error "Top-level declaration not found."
            i = case find_instance_of_ident_in_decl ident d of
                  Just res -> res 
                  Nothing -> error "No instance found."
        RefacAlphaEq.replaceIdent_API i modname target_ident